""" batch functions """
from __future__ import annotations

import logging
from typing import Optional

from nowcasting_dataset.config.model import Configuration
from nowcasting_dataset.dataset.batch import Batch
from pydantic import BaseModel

from nowcasting_dataloader.data_sources import (
    GSPML,
    NWPML,
    PVML,
    MetadataML,
    SatelliteML,
    SunML,
    TopographicML,
)
from nowcasting_dataloader.xr_utils import (
    register_xr_data_array_to_tensor,
    register_xr_data_set_to_tensor,
)

_LOG = logging.getLogger(__name__)

register_xr_data_array_to_tensor()
register_xr_data_set_to_tensor()


class Example(BaseModel):
    """
    Single Data item

    Note that this is currently not really used
    """

    satellite: Optional[SatelliteML]
    hrvsatellite: Optional[SatelliteML]
    topographic: Optional[TopographicML]
    pv: Optional[PVML]
    sun: Optional[SunML]
    gsp: Optional[GSPML]
    nwp: Optional[NWPML]

    @property
    def data_sources(self):
        """The different data sources"""
        return [
            self.satellite,
            self.hrvsatellite,
            self.topographic,
            self.pv,
            self.sun,
            self.gsp,
            self.nwp,
        ]


class BatchML(Example):
    """
    Batch data object.

    Contains the following data sources
    - gsp, satellite, topogrpahic, sun, pv, nwp and datetime.
    Also contains metadata of the class

    """

    metadata: MetadataML

    @staticmethod
    def fake(configuration: Configuration = Configuration()):
        """Create fake batch"""

        batch: Batch = Batch.fake(configuration=configuration)
        batch: BatchML = BatchML.from_batch(batch=batch)

        return batch

    @staticmethod
    def from_batch(batch: Batch) -> BatchML:
        """Change batch to ML batch"""
        data_sources_names = Example.__fields__.keys()

        data_sources_dict = {}
        for data_source_name in data_sources_names:

            data_source = BatchML.__fields__[data_source_name].type_

            xr_dataset = getattr(batch, data_source_name)
            if xr_dataset is not None:
                try:
                    data_sources_dict[data_source_name] = data_source.from_xr_dataset(xr_dataset)
                except Exception as e:
                    _LOG.error(
                        f"Could not change xr dataset to " f"pydantic model for {data_source_name}"
                    )
                    raise e

                if "satellite" in data_source_name or "nwp" in data_source_name:
                    # Add in the channels being used
                    # Only need it from the first example
                    data_sources_dict[data_source_name].channels = xr_dataset["channels"][0].values

        data_sources_dict["metadata"] = batch.metadata.dict()

        return BatchML(**data_sources_dict)

    def normalize(self):
        """Normalize the batch"""

        # loop over all data sources and normalize
        for data_sources in self.data_sources:
            if data_sources is not None:
                data_sources.normalize()



""" General Data Source output pydantic class. """
from __future__ import annotations

import logging
from typing import List, Union

import numpy as np
import torch
import xarray as xr
from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)

Array = Union[xr.DataArray, np.ndarray, torch.Tensor, list]

OSGB_X_MAX = 654665
OSGB_Y_MAX = 1151577


class DataSourceOutputML(BaseModel):
    """General Data Source output pydantic class.

    Data source output classes should inherit from this class
    """

    class Config:
        """Allowed classes e.g. tensor.Tensor"""

        # TODO maybe there is a better way to do this
        arbitrary_types_allowed = True

    batch_size: int = Field(
        0,
        ge=0,
        description="The size of this batch. If the batch size is 0, "
        "then this item stores one data item i.e Example",
    )

    normalized: bool = Field(
        False,
        description="If the data field has been normalized or not",
    )

    def get_name(self) -> str:
        """Get the name of the class"""
        return self.__class__.__name__.lower()

    def get_datetime_index(self):
        """Datetime index for the data"""
        pass

    def normalize(self):
        """Normalize the data"""
        pass


def pad_nans(array, pad_width) -> np.ndarray:
    """Pad nans with nans"""
    array = array.astype(np.float32)
    return np.pad(array, pad_width, constant_values=np.NaN)


def pad_data(
    data: DataSourceOutputML,
    pad_size: int,
    one_dimensional_arrays: List[str],
    two_dimensional_arrays: List[str],
):
    """
    Pad (if necessary) so returned arrays are always of size

    data has two types of arrays in it, one dimensional arrays and two dimensional arrays
    the one dimensional arrays are padded in that dimension
    the two dimensional arrays are padded in the second dimension

    Note that class is edited so nothing is returned.

    Args:
        data: typed dictionary of data objects
        pad_size: the maount that should be padded
        one_dimensional_arrays: list of data items that should be padded by one dimension
        two_dimensional_arrays: list of data tiems that should be padded in the
            third dimension (and more)

    """
    # Pad (if necessary) so returned arrays are always of size
    pad_shape = (0, pad_size)  # (before, after)

    for name in one_dimensional_arrays:
        data.__setattr__(name, pad_nans(data.__getattribute__(name), pad_width=pad_shape))

    for variable in two_dimensional_arrays:
        data.__setattr__(
            variable, pad_nans(data.__getattribute__(variable), pad_width=((0, 0), pad_shape))
        )  # (axis0, axis1)



""" Dataset and functions"""
import logging
import os
from typing import List, Optional, Tuple, Union

import boto3
import einops
import gcsfs
import torch
from nowcasting_dataset.config.model import Configuration
from nowcasting_dataset.consts import (
    DEFAULT_REQUIRED_KEYS,
    GSP_DATETIME_INDEX,
    GSP_ID,
    GSP_YIELD,
    NWP_DATA,
    PV_SYSTEM_ID,
    PV_YIELD,
    SATELLITE_DATA,
    SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME,
    TOPOGRAPHIC_DATA,
)
from nowcasting_dataset.dataset.batch import Batch, Example
from nowcasting_dataset.filesystem.utils import delete_all_files_in_temp_path, download_to_local
from nowcasting_dataset.utils import get_netcdf_filename, set_fsspec_for_multiprocess

from nowcasting_dataloader.batch import BatchML
from nowcasting_dataloader.subset import subselect_data
from nowcasting_dataloader.utils.position_encoding import generate_position_encodings_for_batch

logger = logging.getLogger(__name__)


class NetCDFDataset(torch.utils.data.Dataset):
    """
    Loads data saved by the `prepare_ml_training_data.py` script.

    Moved from predict_pv_yield
    """

    def __init__(
        self,
        n_batches: int,
        src_path: str,
        tmp_path: str,
        configuration: Configuration,
        cloud: str = "gcp",
        required_keys: Union[Tuple[str], List[str]] = None,
        history_minutes: Optional[int] = None,
        forecast_minutes: Optional[int] = None,
        normalize: bool = True,
        add_position_encoding: bool = False,
        data_sources_names: Optional[list[str]] = None,
        num_bands: int = 4,
    ):
        """
        Netcdf Dataset

        Args:
            n_batches: Number of batches available on disk.
            src_path: The full path (including 'gs://') to the data on
                Google Cloud storage.
            tmp_path: The full path to the local temporary directory
                (on a local filesystem).
            cloud:
            required_keys: Tuple or list of keys required in the example for
                it to be considered usable
            history_minutes: How many past minutes of data to use, if subsetting the batch
            forecast_minutes: How many future minutes of data to use, if reducing the amount
                of forecast time
            configuration: configuration object
            cloud: which cloud is used, can be "gcp", "aws" or "local".
            normalize: normalize the batch data
            add_position_encoding: Whether to add position encoding or not
            data_sources_names: Names of data sources to load, if not using all of them
            num_bands: Number of bands for the Fourier features for the position encoding
        """
        self.n_batches = n_batches
        self.src_path = src_path
        self.tmp_path = tmp_path
        self.cloud = cloud
        self.history_minutes = history_minutes
        self.forecast_minutes = forecast_minutes
        self.configuration = configuration
        self.normalize = normalize
        self.add_position_encoding = add_position_encoding
        self.num_bands = num_bands
        if data_sources_names is None:
            data_sources_names = list(Example.__fields__.keys())
        self.data_sources_names = data_sources_names

        logger.info(f"Setting up NetCDFDataset for {src_path}")

        if self.forecast_minutes is None:
            self.forecast_minutes = configuration.input_data.default_forecast_minutes
        if self.history_minutes is None:
            self.history_minutes = configuration.input_data.default_history_minutes

        # see if we need to select the subset of data. If turned on -
        # only history_minutes + current time + forecast_minutes data is used.
        self.select_subset_data = False
        if self.forecast_minutes != configuration.input_data.default_forecast_minutes:
            self.select_subset_data = True
        if self.history_minutes != configuration.input_data.default_history_minutes:
            self.select_subset_data = True

        # Index into either sat_datetime_index or nwp_target_time indicating the current time,
        self.current_timestep_5_index = (
            int(configuration.input_data.default_history_minutes // 5) + 1
        )

        if required_keys is None:
            required_keys = DEFAULT_REQUIRED_KEYS
        self.required_keys = list(required_keys)

        # setup cloud connections as None
        self.gcs = None
        self.s3_resource = None

        assert cloud in ["gcp", "aws", "local"]

        if not os.path.isdir(self.tmp_path):
            os.mkdir(self.tmp_path)

    def per_worker_init(self, worker_id: int):
        """Function called by a worker"""
        if self.cloud == "gcp":
            self.gcs = gcsfs.GCSFileSystem()
        elif self.cloud == "aws":
            self.s3_resource = boto3.resource("s3")

        # adjust temp path for each worker
        self.tmp_path = f"{self.tmp_path}/{worker_id}"

    def __len__(self):
        """Length of dataset"""
        return self.n_batches

    def __getitem__(self, batch_idx: int) -> dict:
        """Returns a whole batch at once.

        Args:
          batch_idx: The integer index of the batch. Must be in the range
          [0, self.n_batches).

        Returns:
            NamedDict where each value is a numpy array. The size of this
            array's first dimension is the batch size.
        """
        logger.debug(f"Getting batch {batch_idx}")
        if not 0 <= batch_idx < self.n_batches:
            raise IndexError(
                "batch_idx must be in the range" f" [0, {self.n_batches}), not {batch_idx}!"
            )

        if self.cloud in ["gcp", "aws"]:
            # download all data files
            for data_source in self.data_sources_names:
                data_source_and_filename = f"{data_source}/{get_netcdf_filename(batch_idx)}"
                download_to_local(
                    remote_filename=f"{self.src_path}/{data_source_and_filename}",
                    local_filename=f"{self.tmp_path}/{data_source_and_filename}",
                )

            # download locations file
            download_to_local(
                remote_filename=f"{self.src_path}/"
                f"{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME}",
                local_filename=f"{self.tmp_path}/"
                f"{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME}",
            )

            local_netcdf_folder = self.tmp_path
        else:
            local_netcdf_folder = self.src_path

        batch: Batch = Batch.load_netcdf(
            local_netcdf_folder, batch_idx=batch_idx, data_sources_names=self.data_sources_names
        )

        if self.select_subset_data:
            batch = subselect_data(
                batch=batch,
                history_minutes=self.history_minutes,
                forecast_minutes=self.forecast_minutes,
                current_timestep_index=self.current_timestep_5_index,
            )
        if self.add_position_encoding:
            position_encodings = generate_position_encodings_for_batch(
                batch, num_bands=self.num_bands
            )
        # change batch into ML learning batch ready for training
        try:
            batch: BatchML = BatchML.from_batch(batch=batch)
        except Exception as e:
            logger.error(
                f"Could not change Batch to BatchML " f"for batch index {batch_idx}, {batch}"
            )
            raise e

        if self.cloud != "local":
            # remove files in a folder, but not the folder itself
            delete_all_files_in_temp_path(self.tmp_path)

        # normalize the data
        if self.normalize:
            batch.normalize()

        batch: dict = batch.dict()
        if self.add_position_encoding:
            # Add position encodings
            batch.update(position_encodings)
        return batch


class SatFlowDataset(NetCDFDataset):
    """SatFlow dataset for filtering and splitting up output from NetCDFDataset properly"""

    def __init__(
        self,
        n_batches: int,
        src_path: str,
        tmp_path: str,
        configuration: Configuration,
        cloud: str = "gcp",
        required_keys: Union[Tuple[str], List[str]] = None,
        history_minutes: Optional[int] = None,
        forecast_minutes: Optional[int] = None,
        normalize: bool = True,
        add_position_encoding: bool = False,
        add_satellite_target: bool = False,
        add_hrv_satellite_target: bool = False,
        data_sources_names: Optional[list[str]] = None,
    ):
        """
        Netcdf Dataset

        Args:
            n_batches: Number of batches available on disk.
            src_path: The full path (including 'gs://') to the data on
                Google Cloud storage.
            tmp_path: The full path to the local temporary directory
                (on a local filesystem).
            cloud:
            required_keys: Tuple or list of keys required in the example for
                it to be considered usable
            history_minutes: How many past minutes of data to use, if subsetting the batch
            forecast_minutes: How many future minutes of data to use, if reducing the amount
                of forecast time
            configuration: configuration object
            cloud: which cloud is used, can be "gcp", "aws" or "local".
            normalize: normalize the batch data
            add_position_encoding: Whether to add position encoding or not
            add_satellite_target: Whether to add future satellite imagery to the target or not
            add_hrv_satellite_target: Whether to add future HRV satellite imagery to the target
            data_sources_names: Names of data sources to load, if not using all of them
        """
        super().__init__(
            n_batches=n_batches,
            src_path=src_path,
            tmp_path=tmp_path,
            configuration=configuration,
            cloud=cloud,
            required_keys=required_keys,
            history_minutes=history_minutes,
            forecast_minutes=forecast_minutes,
            normalize=normalize,
            add_position_encoding=add_position_encoding,
            data_sources_names=data_sources_names,
        )

        self.add_satellite_target = add_satellite_target
        self.add_hrv_satellite_target = add_hrv_satellite_target

        # SatFlow specific changes, i.e. which timestep to split on
        self.current_timestep_index = (self.history_minutes // 5) + 1
        self.current_timestep_index_30 = (self.history_minutes // 30) + 1

    def __getitem__(self, batch_idx: int) -> Tuple[dict, dict]:
        """
        Satflow extension for the dataloader

        Args:
            batch_idx: Batch ID to load

        Returns:
            Tuple of dicts of torch.Tensors holding the data, with the first one being the inputs
            and the second being the target
        """
        batch = super().__getitem__(batch_idx)
        x = {}
        target = {}
        # Need to partition out past and future sat images here, along with the rest of the data
        if "satellite" in self.data_sources_names and len(batch["satellite"].get("data", [])) > 0:
            past_satellite_data = batch["satellite"]["data"][:, :, : self.current_timestep_index]
            x["satellite"] = past_satellite_data
        if (
            "hrvsatellite" in self.data_sources_names
            and len(batch["hrvsatellite"].get("data", [])) > 0
        ):
            past_hrv_satellite_data = batch["hrvsatellite"]["data"][
                :, :, : self.current_timestep_index
            ]
            x["hrvsatellite"] = past_hrv_satellite_data
        if "pv" in self.data_sources_names and len(batch["pv"].get(PV_YIELD, [])) > 0:
            past_pv_data = torch.unsqueeze(
                batch["pv"][PV_YIELD][:, : self.current_timestep_index], dim=1
            )
            x[PV_YIELD] = past_pv_data
            x[PV_SYSTEM_ID] = torch.nan_to_num(batch["pv"][PV_SYSTEM_ID])
        if "nwp" in self.data_sources_names and len(batch["nwp"].get("data", [])) > 0:
            # We can give future NWP too, as that will be available
            x[NWP_DATA] = batch["nwp"]["data"]
        if (
            "topographic" in self.data_sources_names
            and len(batch["topographic"].get(TOPOGRAPHIC_DATA, [])) > 0
        ):
            # Need to expand dims to get a single channel one
            # Results in topographic maps with [Batch, Channel, H, W]
            x[TOPOGRAPHIC_DATA] = torch.unsqueeze(
                torch.unsqueeze(batch["topographic"][TOPOGRAPHIC_DATA], dim=1), dim=1
            )

        if "gsp" in self.data_sources_names:
            # Only GSP information we give to the model to train on is the IDs & physical locations
            x[GSP_ID] = torch.nan_to_num(batch["gsp"][GSP_ID])

            # Now creating the target data, only want the first GSP as the target
            target[GSP_YIELD] = batch["gsp"][GSP_YIELD][:, self.current_timestep_index_30 :, 0]
            target[GSP_ID] = batch["gsp"][GSP_ID][:, 0]
            # Add timestep, so we can compare results better
            target[GSP_DATETIME_INDEX] = batch["gsp"][GSP_DATETIME_INDEX][
                :, self.current_timestep_index_30 - 1 :
            ]
            target["gsp_capacity"] = batch["gsp"]["gsp_capacity"][
                :, self.current_timestep_index_30 :, 0
            ]

        if self.add_satellite_target:
            future_sat_data = batch["satellite"]["data"][:, :, self.current_timestep_index :]
            target[SATELLITE_DATA] = future_sat_data
        if self.add_hrv_satellite_target:
            future_hrv_sat_data = batch["hrvsatellite"]["data"][:, :, self.current_timestep_index :]
            target["hrv_" + SATELLITE_DATA] = future_hrv_sat_data

        # Add position encodings
        if self.add_position_encoding:
            if len(x.get("satellite", [])) > 0:
                x = self.add_encodings(
                    x, "satellite", batch, self.current_timestep_index, self.add_satellite_target
                )
                if self.add_hrv_satellite_target:
                    x[SATELLITE_DATA + "_query"] = x.pop("satellite_query")
            if len(x.get("hrvsatellite", [])) > 0:
                x = self.add_encodings(
                    x,
                    "hrvsatellite",
                    batch,
                    self.current_timestep_index,
                    self.add_hrv_satellite_target,
                )
                if self.add_hrv_satellite_target:
                    x["hrv_" + SATELLITE_DATA + "_query"] = x.pop("hrvsatellite_query")
            if len(x.get(TOPOGRAPHIC_DATA, [])) > 0:
                x[TOPOGRAPHIC_DATA] = torch.cat(
                    [x[TOPOGRAPHIC_DATA], batch["topographic_position_encoding"]], dim=1
                )
            if len(x.get(NWP_DATA, [])) > 0:
                x[NWP_DATA] = torch.cat(
                    [x[NWP_DATA], batch[NWP_DATA + "_position_encoding"]], dim=1
                )
            if len(x.get(PV_YIELD, [])) > 0:
                past_encoding = batch["pv_position_encoding"][:, :, : self.current_timestep_index]
                x[PV_YIELD] = torch.cat([x[PV_YIELD], past_encoding], dim=1)

            if "gsp" in self.data_sources_names:
                # Add the future GSP position encoding for querying
                x[GSP_YIELD + "_query"] = batch["gsp_position_encoding"][
                    :, :, self.current_timestep_index_30 :
                ]

        # Rename to match other ones better
        if len(x.get("satellite", [])) > 0:
            x[SATELLITE_DATA] = x.pop("satellite")
        if len(x.get("hrvsatellite", [])) > 0:
            x["hrv_" + SATELLITE_DATA] = x.pop("hrvsatellite")

        # Zero out NaN PV and GSP Yield
        x = self.zero_out_nan_pv_systems(x)

        # Convert all to float32 if not already done so
        for k in x.keys():
            if x[k].dtype != torch.float32:
                x[k] = x[k].float()
        for k in target.keys():
            if target[k].dtype != torch.float32:
                target[k] = target[k].float()

        return x, target

    def add_encodings(
        self,
        x: dict,
        key: str,
        batch: dict,
        current_timestep_index: int,
        add_future_encodings: bool = False,
    ) -> dict[str, torch.Tensor]:
        """
        Adds encodings to the targets and inputs

        Args:
            x: Dictionary containing model inputs
            key: Key to check and insert
            batch: Batch of data to use
            current_timestep_index: The index of the t0 time in the data, used to split into past
                and future encodings
            add_future_encodings: Whether to add the future position encodings to the target

        Returns:
            x dictionary with the added/updated keys
        """

        if key + "_position_encoding" in batch:
            past_encoding = batch[key + "_position_encoding"][:, :, :current_timestep_index]
            x[key] = torch.cat([x[key], past_encoding], dim=1)
            if add_future_encodings:
                future_encoding = batch[key + "_position_encoding"][:, :, current_timestep_index:]
                x[key + "_query"] = future_encoding
        return x

    def zero_out_nan_pv_systems(self, x: dict) -> dict:
        """
        Zeros out NaN PV systems and their position encodings

        This takes advantage of the fact that NaN values for PV and GSP systems are always at the
        end of the example, so just need to find the first one and can zero out everything after.

        This should be used after the position encodings are appended, so that it zeros out all the
        channels as well

        Args:
            x: dictionary of inputs

        Returns:
            The dictionary x with the PV systems zeroed out and position encodings zeroed as well
        """

        for key in [PV_YIELD, GSP_YIELD]:
            if key in x:
                mask = torch.isnan(x[key][:, 0, 0, :])  # Only looking at the yield
                mask = einops.repeat(mask, "b id -> b c t id", c=x[key].shape[1], t=x[key].shape[2])
                # Zero out for all entries related to
                x[key][mask] = 0.0

        return x


def worker_init_fn(worker_id):
    """Configures each dataset worker process.

    1. Get fsspec ready for multi process
    2. To call NowcastingDataset.per_worker_init().
    """
    # fix for fsspec when using multprocess
    set_fsspec_for_multiprocess()

    # get_worker_info() returns information specific to each worker process.
    worker_info = torch.utils.data.get_worker_info()
    if worker_info is None:
        print("worker_info is None!")
    else:
        # The NowcastingDataset copy in this worker process.
        dataset_obj = worker_info.dataset
        dataset_obj.per_worker_init(worker_info.id)



""" Useful functions for xarray objects

1. xr array and xr dataset --> to torch functions
"""
from typing import List

import numpy as np
import torch
import xarray as xr


def map_channels_names_to_indexes(xr_dataset: xr.Dataset, channels_mapping: dict):
    """
    Map channel names to indexes

    xr_dataset.channels = [['HRV', 'IR_016'],
       ['HRV', 'IR_016'],
       ['HRV', 'IR_016'],
       ['HRV', 'IR_016']]

    will be changed to

    xr_dataset.channels = [[0, 1],
       [0, 1],
       [0, 1],
       [0, 1]]

    """
    # map channels to indexes
    mapping_np = np.vectorize(channels_mapping.__getitem__)(xr_dataset.channels.values)

    # make data array
    channels_xr = xr.DataArray(mapping_np, dims=xr_dataset.channels.dims)

    # set data array
    xr_dataset.__setitem__("channels", channels_xr)

    return xr_dataset


def register_xr_data_array_to_tensor():
    """Add torch object to data array"""
    if not hasattr(xr.DataArray, "torch"):

        @xr.register_dataarray_accessor("torch")
        class TorchAccessor:
            def __init__(self, xarray_obj):
                self._obj = xarray_obj

            def to_tensor(self):
                """Convert this DataArray to a torch.Tensor"""
                return torch.tensor(self._obj.data, dtype=torch.float32)

            # torch tensor names does not working in dataloader yet - 2021-10-15
            # https://discuss.pytorch.org/t/collating-named-tensors/78650
            # https://github.com/openclimatefix/nowcasting_dataset/issues/25
            # def to_named_tensor(self):
            #     """Convert this DataArray to a torch.Tensor with named dimensions"""
            #     import torch
            #
            #     return torch.tensor(self._obj.data, names=self._obj.dims)


def register_xr_data_set_to_tensor():
    """Add torch object to dataset"""
    if not hasattr(xr.Dataset, "torch"):

        @xr.register_dataset_accessor("torch")
        class TorchAccessor:
            def __init__(self, xdataset_obj: xr.Dataset):
                self._obj = xdataset_obj

            def to_tensor(self, data_vars: List[str]) -> dict:
                """Convert this Dataset to dictionary of torch tensors"""
                torch_dict = {}

                for data_var in data_vars:
                    v = getattr(self._obj, data_var)

                    if data_var.find("time") != -1:
                        time_int = v.data.astype(int)
                        torch_dict[data_var] = torch.tensor(time_int, dtype=torch.float64)

                    else:
                        torch_dict[data_var] = v.torch.to_tensor()

                return torch_dict


register_xr_data_array_to_tensor()
register_xr_data_set_to_tensor()


def re_order_dims(xr_dataset: xr.Dataset):
    """
    Re order dims to B,C,T,H,W
    """
    expected_dims_order = ("example", "channels_index", "time_index", "y_index", "x_index")

    if xr_dataset.data.dims != expected_dims_order:
        xr_dataset.__setitem__("data", xr_dataset.data.transpose(*expected_dims_order))

    return xr_dataset



""" Model for output of PV data """
import logging

import numpy as np
from nowcasting_dataset.consts import (
    PV_DATETIME_INDEX,
    PV_SYSTEM_ID,
    PV_SYSTEM_ROW_NUMBER,
    PV_SYSTEM_X_COORDS,
    PV_SYSTEM_Y_COORDS,
    PV_YIELD,
)
from pydantic import Field, validator

from nowcasting_dataloader.data_sources.datasource_output import (
    OSGB_X_MAX,
    OSGB_Y_MAX,
    Array,
    DataSourceOutputML,
)

logger = logging.getLogger(__name__)


class PVML(DataSourceOutputML):
    """Model for output of PV data"""

    # Shape: [batch_size,] seq_length, n_pv_systems_per_example
    pv_yield: Array = Field(
        ...,
        description=" PV yield from all PV systems in the region of interest (ROI). \
    : Includes central PV system, which will always be the first entry. \
    : shape = [batch_size, ] seq_length, n_pv_systems_per_example",
    )

    # Shape: [batch_size,], n_pv_systems_per_example
    pv_capacity: Array = Field(
        ...,
        description=" PV capacity from all PV systems in the region of interest (ROI). \
        : Includes central PV system, which will always be the first entry. "
        "PV capacity is assume constant: \
        : shape = [batch_size, ], n_pv_systems_per_example",
    )

    #: PV identification.
    #: shape = [batch_size, ] n_pv_systems_per_example
    pv_system_id: Array = Field(..., description="PV system ID, e.g. from PVoutput.org")
    pv_system_row_number: Array = Field(..., description="pv row number, made by OCF  TODO")

    pv_datetime_index: Array = Field(
        ...,
        description="The datetime associated with the pv system data. "
        "shape = [batch_size, ] sequence length,",
    )

    pv_system_x_coords: Array = Field(
        ...,
        description="The x (OSGB geo-spatial) coordinates of the pv systems. "
        "Shape: [batch_size,] n_pv_systems_per_example",
    )
    pv_system_y_coords: Array = Field(
        ...,
        description="The y (OSGB geo-spatial) coordinates of the pv systems. "
        "Shape: [batch_size,] n_pv_systems_per_example",
    )

    @property
    def number_of_pv_systems(self):
        """The number of pv systems"""
        return self.pv_yield.shape[-1]

    @property
    def sequence_length(self):
        """The sequence length of the pv data"""
        return self.pv_yield.shape[-2]

    @validator("pv_system_x_coords")
    def x_coordinates_shape(cls, v, values):
        """Validate 'pv_system_x_coords'"""
        assert v.shape[-1] == values["pv_yield"].shape[-1]

        if values["normalized"]:
            if v.max() > 1:
                raise Exception(
                    f"PV data is normalized, "
                    f"but PV X coordinates maximum value is above 1, {v.max()}"
                )
            if v.min() < 0:
                raise Exception(
                    f"PV data is normalized, "
                    f"but PV X coordinates minimum value is below 0, {v.min()}"
                )

        return v

    @validator("pv_system_y_coords")
    def y_coordinates(cls, v, values):
        """Validate 'pv_system_y_coords'"""
        assert v.shape[-1] == values["pv_yield"].shape[-1]

        if values["normalized"]:
            if v.max() > 1:
                raise Exception(
                    f"PV data is normalized, "
                    f"but PV Y coordinates maximum value is above 1, {v.max()}"
                )
            if v.min() < 0:
                raise Exception(
                    f"PV data is normalized, "
                    f"but PV Y coordinates minimum value is below 0, {v.min()}"
                )

        return v

    def get_datetime_index(self) -> Array:
        """Get the datetime index of this data"""
        return self.pv_datetime_index

    @staticmethod
    def from_xr_dataset(xr_dataset):
        """Change xr dataset to model. If data does not exist, then return None"""

        pv_batch_ml = xr_dataset.torch.to_tensor(
            ["power_mw", "capacity_mwp", "time", "x_coords", "y_coords", "id"]
        )

        pv_batch_ml[PV_YIELD] = pv_batch_ml.pop("power_mw")
        pv_batch_ml["pv_capacity"] = pv_batch_ml.pop("capacity_mwp")
        pv_batch_ml[PV_SYSTEM_ID] = pv_batch_ml["id"]
        pv_batch_ml[PV_SYSTEM_ROW_NUMBER] = pv_batch_ml.pop("id")
        pv_batch_ml[PV_DATETIME_INDEX] = pv_batch_ml.pop("time")
        pv_batch_ml[PV_SYSTEM_X_COORDS] = pv_batch_ml.pop("x_coords")
        pv_batch_ml[PV_SYSTEM_Y_COORDS] = pv_batch_ml.pop("y_coords")

        return PVML(**pv_batch_ml)

    def normalize(self):
        """Normalize the gsp data"""
        if not self.normalized:
            # Expand capacity to the same timesteps for broadcasting
            capacity = np.expand_dims(self.pv_capacity, axis=1)
            self.pv_yield = self.pv_yield / capacity

            self.pv_system_x_coords = self.pv_system_x_coords / OSGB_X_MAX
            self.pv_system_y_coords = self.pv_system_y_coords / OSGB_Y_MAX

            self.normalized = True



""" Model for output of GSP data """
import logging

from nowcasting_dataset.consts import (
    GSP_DATETIME_INDEX,
    GSP_ID,
    GSP_X_COORDS,
    GSP_Y_COORDS,
    GSP_YIELD,
)
from pydantic import Field, validator

from nowcasting_dataloader.data_sources.datasource_output import (
    OSGB_X_MAX,
    OSGB_Y_MAX,
    Array,
    DataSourceOutputML,
)

logger = logging.getLogger(__name__)


class GSPML(DataSourceOutputML):
    """Model for output of GSP data"""

    # Shape: [batch_size,] seq_length, n_pv_systems_per_example
    gsp_yield: Array = Field(
        ...,
        description=" GSP yield from all GSP in the region of interest (ROI). \
    : Includes central GSP system, which will always be the first entry. \
    : shape = [batch_size, ] seq_length, n_gsp_per_example",
    )

    # Shape: [batch_size,] seq_length, n_pv_systems_per_example
    gsp_capacity: Array = Field(
        ...,
        description=" GSP capacity from all GSP in the region of interest (ROI). \
        : Includes central GSP system, which will always be the first entry. \
        : shape = [batch_size, ] seq_length, n_gsp_per_example",
    )

    #: GSP identification.
    #: shape = [batch_size, ] n_pv_systems_per_example
    gsp_id: Array = Field(..., description="gsp id from NG")

    gsp_datetime_index: Array = Field(
        ...,
        description="The datetime associated with the gsp data. "
        "shape = [batch_size, ] sequence length,",
    )

    gsp_x_coords: Array = Field(
        ...,
        description="The x (OSGB geo-spatial) coordinates of the gsp. "
        "This is in fact the x centroid of the GSP region"
        "Shape: [batch_size,] n_gsp_per_example",
    )
    gsp_y_coords: Array = Field(
        ...,
        description="The y (OSGB geo-spatial) coordinates of the gsp. "
        "This are in fact the y centroid of the GSP region"
        "Shape: [batch_size,] n_gsp_per_example",
    )

    @property
    def number_of_gsp(self):
        """The number of Grid Supply Points in this example"""
        return self.gsp_yield.shape[-1]

    @property
    def sequence_length(self):
        """The sequence length of the GSP PV power timeseries data"""
        return self.gsp_yield.shape[-2]

    @validator("gsp_yield")
    def gsp_yield_shape(cls, v, values):
        """Validate 'gsp_yield'"""
        assert len(v.shape) == 3

        return v

    @validator("gsp_x_coords")
    def x_coordinates(cls, v, values):
        """Validate 'gsp_x_coords'"""
        assert v.shape[-1] == values["gsp_yield"].shape[-1]
        if values["normalized"]:
            if v.max() > 1:
                raise Exception(
                    f"GSP data is normalized, "
                    f"but GSP X coordinates maximum value is above 1, {v.max()}"
                )
            if v.min() < 0:
                raise Exception(
                    f"GSP data is normalized, "
                    f"but GSP X coordinates minimum value is below 0, {v.min()}"
                )
        return v

    @validator("gsp_y_coords")
    def y_coordinates_shape(cls, v, values):
        """Validate 'gsp_y_coords'"""
        assert v.shape[-1] == values["gsp_yield"].shape[-1]
        if values["normalized"]:
            if v.max() > 1:
                raise Exception(
                    f"GSP data is normalized, "
                    f"but GSP Y coordinates maximum value is above 1, {v.max()}"
                )
            if v.min() < 0:
                raise Exception(
                    f"GSP data is normalized, "
                    f"but GSP Y coordinates minimum value is below 0, {v.min()}"
                )
        return v

    def get_datetime_index(self) -> Array:
        """Get the datetime index of this data"""
        return self.gsp_datetime_index

    @staticmethod
    def from_xr_dataset(xr_dataset):
        """Change xr dataset to model. If data does not exist, then return None"""

        gsp_batch_ml = xr_dataset.torch.to_tensor(
            ["power_mw", "capacity_mwp", "time", "x_coords", "y_coords", "id"]
        )

        gsp_batch_ml[GSP_YIELD] = gsp_batch_ml.pop("power_mw")
        gsp_batch_ml["gsp_capacity"] = gsp_batch_ml.pop("capacity_mwp")
        gsp_batch_ml[GSP_ID] = gsp_batch_ml.pop("id")
        gsp_batch_ml[GSP_DATETIME_INDEX] = gsp_batch_ml.pop("time")
        gsp_batch_ml[GSP_X_COORDS] = gsp_batch_ml.pop("x_coords")
        gsp_batch_ml[GSP_Y_COORDS] = gsp_batch_ml.pop("y_coords")
        gsp_batch_ml["batch_size"] = gsp_batch_ml[GSP_YIELD].shape[0]

        return GSPML(**gsp_batch_ml)

    def normalize(self):
        """Normalize the gsp data"""
        if not self.normalized:
            self.gsp_yield = self.gsp_yield / self.gsp_capacity

            self.gsp_x_coords = self.gsp_x_coords / OSGB_X_MAX
            self.gsp_y_coords = self.gsp_y_coords / OSGB_Y_MAX

            self.normalized = True



""" Various DataSources """
from .gsp.gsp_model import GSPML
from .metadata.metadata_model import MetadataML
from .nwp.nwp_model import NWPML
from .pv.pv_model import PVML
from .satellite.satellite_model import SatelliteML
from .sun.sun_model import SunML
from .topographic.topographic_model import TopographicML



"""Datamodules to use for training"""
import logging
import os
from typing import List, Optional, Tuple, Union

import torch
from nowcasting_dataset.config import load_yaml_configuration
from nowcasting_dataset.config.model import Configuration
from pytorch_lightning import LightningDataModule

from nowcasting_dataloader.datasets import NetCDFDataset, SatFlowDataset, worker_init_fn
from nowcasting_dataloader.fake import FakeDataset

_LOG = logging.getLogger(__name__)


class SatFlowDataModule(LightningDataModule):
    """
    Example of LightningDataModule for SatFlow dataset.

    A DataModule implements 5 key methods:
        - prepare_data (things to do on 1 GPU/TPU, not on every GPU/TPU in distributed mode)
        - setup (things to do on every accelerator in distributed mode)
        - train_dataloader (the training dataloader)
        - val_dataloader (the validation dataloader(s))
        - test_dataloader (the test dataloader(s))
    This allows you to share a full dataset without explaining how to download,
    split, transform and process the data.
    Read the docs:
        https://pytorch-lightning.readthedocs.io/en/latest/extensions/datamodules.html
    """

    def __init__(
        self,
        temp_path: str,
        configuration: Union[Configuration, str],
        cloud: str = "local",
        required_keys: Union[Tuple[str], List[str]] = None,
        history_minutes: Optional[int] = None,
        forecast_minutes: Optional[int] = None,
        normalize: bool = True,
        add_position_encoding: bool = False,
        add_satellite_target: bool = False,
        add_hrv_satellite_target: bool = False,
        data_sources_names: Optional[list[str]] = None,
        pin_memory: bool = True,
        num_workers: int = 1,
    ):
        """
        Datamodule for the SatFlow dataset

        Args:
            temp_path: temp path of data
            configuration: Configuration to use, or path to configuration file
            cloud: What cloud to use, defaults to local
            required_keys: Required keys for the dataset
            history_minutes: Number of history minutes to use
            forecast_minutes: Number of forecast minutes to use
            normalize: Whether to normalize the data
            add_position_encoding: Whether to add position encoding
            add_satellite_target: Whether to add satellite imagery to the targets
            add_hrv_satellite_target: Whether to add HRV satellite target
            data_sources_names: Names of data sources to load. Leave as None to load all.
            pin_memory: Whether to pin memory in the dataloader
            num_workers: Number of workers for each dataloader
        """
        super().__init__()
        self.temp_path = temp_path
        if type(configuration) == str:
            configuration = load_yaml_configuration(configuration)
        self.configuration: Configuration = configuration
        self.cloud = cloud
        self.n_train_data = self.configuration.process.n_train_batches
        self.n_val_data = self.configuration.process.n_validation_batches
        self.n_test_data = self.configuration.process.n_test_batches
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.required_keys = required_keys
        self.forecast_minutes = forecast_minutes
        self.history_minutes = history_minutes
        self.normalize = normalize
        self.add_position_encoding = add_position_encoding
        self.add_satellite_target = add_satellite_target
        self.add_hrv_satellite_target = add_hrv_satellite_target
        self.data_sources_names = data_sources_names

        self.dataloader_config = dict(
            pin_memory=self.pin_memory,
            num_workers=self.num_workers,
            prefetch_factor=8,
            worker_init_fn=worker_init_fn,
            persistent_workers=True,
            # Disable automatic batching because dataset
            # returns complete batches.
            batch_size=None,
        )

    def train_dataloader(self):
        """Train dataloader"""
        train_dataset = SatFlowDataset(
            self.n_train_data,
            os.path.join(self.configuration.output_data.filepath, "train"),
            os.path.join(self.temp_path, "train"),
            configuration=self.configuration,
            cloud=self.cloud,
            required_keys=self.required_keys,
            history_minutes=self.history_minutes,
            forecast_minutes=self.forecast_minutes,
            normalize=self.normalize,
            add_position_encoding=self.add_position_encoding,
            add_satellite_target=self.add_satellite_target,
            add_hrv_satellite_target=self.add_hrv_satellite_target,
            data_sources_names=self.data_sources_names,
        )

        return torch.utils.data.DataLoader(train_dataset, shuffle=True, **self.dataloader_config)

    def val_dataloader(self):
        """Validation dataloader"""
        val_dataset = SatFlowDataset(
            self.n_val_data,
            os.path.join(self.configuration.output_data.filepath, "validation"),
            os.path.join(self.temp_path, "validation"),
            configuration=self.configuration,
            cloud=self.cloud,
            required_keys=self.required_keys,
            history_minutes=self.history_minutes,
            forecast_minutes=self.forecast_minutes,
            normalize=self.normalize,
            add_position_encoding=self.add_position_encoding,
            add_satellite_target=self.add_satellite_target,
            add_hrv_satellite_target=self.add_hrv_satellite_target,
            data_sources_names=self.data_sources_names,
        )

        return torch.utils.data.DataLoader(val_dataset, shuffle=False, **self.dataloader_config)

    def test_dataloader(self):
        """Test dataloader"""
        test_dataset = SatFlowDataset(
            self.n_test_data,
            os.path.join(self.configuration.output_data.filepath, "test"),
            os.path.join(self.temp_path, "test"),
            configuration=self.configuration,
            cloud=self.cloud,
            required_keys=self.required_keys,
            history_minutes=self.history_minutes,
            forecast_minutes=self.forecast_minutes,
            normalize=self.normalize,
            add_position_encoding=self.add_position_encoding,
            add_satellite_target=self.add_satellite_target,
            add_hrv_satellite_target=self.add_hrv_satellite_target,
            data_sources_names=self.data_sources_names,
        )

        return torch.utils.data.DataLoader(test_dataset, shuffle=False, **self.dataloader_config)


class NetCDFDataModule(LightningDataModule):
    """
    Example of LightningDataModule for NETCDF dataset.

    A DataModule implements 5 key methods:
        - prepare_data (things to do on 1 GPU/TPU, not on every GPU/TPU in distributed mode)
        - setup (things to do on every accelerator in distributed mode)
        - train_dataloader (the training dataloader)
        - val_dataloader (the validation dataloader(s))
        - test_dataloader (the test dataloader(s))
    This allows you to share a full dataset without explaining how to download,
    split, transform and process the data.
    Read the docs:
        https://pytorch-lightning.readthedocs.io/en/latest/extensions/datamodules.html
    """

    def __init__(
        self,
        temp_path: str = ".",
        n_train_data: int = 24900,
        n_val_data: int = 1000,
        n_test_data: int = 1000,
        cloud: str = "aws",
        num_workers: int = 8,
        pin_memory: bool = True,
        data_path="prepared_ML_training_data/v4/",
        fake_data: bool = False,
        shuffle_train: bool = False,
    ):
        """
        fake_data: random data is created and used instead. This is useful for testing
        """
        super().__init__()

        self.temp_path = temp_path
        self.data_path = data_path
        self.cloud = cloud
        self.n_train_data = n_train_data
        self.n_val_data = n_val_data
        self.n_test_data = n_test_data
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.fake_data = fake_data
        self.shuffle_train = shuffle_train

        filename = os.path.join(data_path, "configuration.yaml")
        _LOG.debug(f"Will be loading the configuration file {filename}")
        self.configuration = load_yaml_configuration(filename=filename)

        self.dataloader_config = dict(
            pin_memory=self.pin_memory,
            num_workers=self.num_workers,
            prefetch_factor=8,
            worker_init_fn=worker_init_fn,
            persistent_workers=True,
            # Disable automatic batching because dataset
            # returns complete batches.
            batch_size=None,
        )

    def train_dataloader(self):
        """Get the train dataloader"""
        if self.fake_data:
            train_dataset = FakeDataset(configuration=self.configuration)
        else:
            train_dataset = NetCDFDataset(
                self.n_train_data,
                os.path.join(self.data_path, "train"),
                os.path.join(self.temp_path, "train"),
                cloud=self.cloud,
                configuration=self.configuration,
            )

        return torch.utils.data.DataLoader(
            train_dataset, shuffle=self.shuffle_train, **self.dataloader_config
        )

    def val_dataloader(self):
        """Get the validation dataloader"""
        if self.fake_data:
            val_dataset = FakeDataset(configuration=self.configuration)
        else:
            val_dataset = NetCDFDataset(
                self.n_val_data,
                os.path.join(self.data_path, "validation"),
                os.path.join(self.temp_path, "validation"),
                cloud=self.cloud,
                configuration=self.configuration,
            )

        return torch.utils.data.DataLoader(val_dataset, shuffle=False, **self.dataloader_config)

    def test_dataloader(self):
        """Get the test dataloader"""
        if self.fake_data:
            test_dataset = FakeDataset(configuration=self.configuration)
        else:
            test_dataset = NetCDFDataset(
                self.n_test_data,
                os.path.join(self.data_path, "test"),
                os.path.join(self.temp_path, "test"),
                cloud=self.cloud,
                configuration=self.configuration,
            )

        return torch.utils.data.DataLoader(test_dataset, shuffle=False, **self.dataloader_config)



""" Model for output of NWP data """
from __future__ import annotations

import logging
from typing import Optional

import numpy as np
import xarray as xr
from pydantic import Field

from nowcasting_dataloader.data_sources.datasource_output import Array, DataSourceOutputML
from nowcasting_dataloader.xr_utils import re_order_dims

logger = logging.getLogger(__name__)

# Means and std computed with
# nowcasting_dataset/scripts/compute_stats_from_batches.py
# using v15 training batches on 2021-11-24.
NWP_MEAN = {
    "t": 285.7799539185846,
    "dswrf": 294.6696933986283,
    "prate": 3.6078121378638696e-05,
    "r": 75.57106712435926,
    "sde": 0.0024915961594965614,
    "si10": 4.931356852411006,
    "vis": 22321.762918384553,
    "lcc": 47.90454236572895,
    "mcc": 44.22781694449808,
    "hcc": 32.87577371914454,
}

NWP_STD = {
    "t": 5.017000766747606,
    "dswrf": 233.1834250473355,
    "prate": 0.00021690701537950742,
    "r": 15.705370079694358,
    "sde": 0.07560040052148084,
    "si10": 2.664583614352396,
    "vis": 12963.802514945439,
    "lcc": 40.06675870700349,
    "mcc": 41.927221148316384,
    "hcc": 39.05157559763763,
}


class NWPML(DataSourceOutputML):
    """Model for output of NWP data"""

    # Shape: [batch_size,] seq_length, width, height, channel
    data: Array = Field(
        ...,
        description=" Numerical weather predictions (NWPs) \
    : Shape: [batch_size,] channel, seq_length, width, height",
    )

    x: Array = Field(
        ...,
        description="The x (OSGB geo-spatial) coordinates of the NWP data. "
        "Shape: [batch_size,] width",
    )
    y: Array = Field(
        ...,
        description="The y (OSGB geo-spatial) coordinates of the NWP data. "
        "Shape: [batch_size,] height",
    )

    time: Array = Field(
        ...,
        description="Time index of nwp data at 5 minutes past the hour {0, 5, ..., 55}. "
        "Datetimes become Unix epochs (UTC) represented as int64 just before being"
        "passed into the ML model.  The 'target time' is the time the NWP is _about_.",
    )

    init_time: Array = Field(..., description="The time when the nwp forecast was made")

    channels: Optional[Array] = Field(None, description="List of the nwp channels")

    def get_datetime_index(self) -> Array:
        """Get the datetime index of this data"""
        return self.target_time

    @staticmethod
    def from_xr_dataset(xr_dataset: xr.Dataset):
        """Change xr dataset to model with tensors"""

        # make sure dims are the in the correct order
        xr_dataset = re_order_dims(xr_dataset)

        # convert to torch dict
        nwp_batch_ml = xr_dataset.torch.to_tensor(["data", "time", "init_time", "x", "y"])

        # make into Model
        return NWPML(**nwp_batch_ml)

    def normalize(self):
        """Normalize the nwp data"""
        if not self.normalized:
            # Only take the channels that are used
            mean = np.array([NWP_MEAN[b] for b in self.channels])
            std = np.array([NWP_STD[b] for b in self.channels])
            # Expand for normaliation
            mean = np.expand_dims(mean, axis=[1, 2, 3])
            std = np.expand_dims(std, axis=[1, 2, 3])
            self.data = self.data - mean
            self.data = self.data / std
            self.normalized = True



""" Model for output of satellite data """
from __future__ import annotations

import logging
from typing import Optional

import numpy as np
import xarray as xr
from pydantic import Field

from nowcasting_dataloader.data_sources.datasource_output import Array, DataSourceOutputML
from nowcasting_dataloader.xr_utils import re_order_dims

logger = logging.getLogger(__name__)


SAT_MEAN = {
    "HRV": 236.13257536395903,
    "IR_016": 291.61620182554185,
    "IR_039": 858.8040610176552,
    "IR_087": 738.3103442750336,
    "IR_097": 773.0910794778366,
    "IR_108": 607.5318145165666,
    "IR_120": 860.6716261423857,
    "IR_134": 925.0477987594331,
    "VIS006": 228.02134593063957,
    "VIS008": 257.56333202381205,
    "WV_062": 633.5975770915588,
    "WV_073": 543.4963868823854,
}

SAT_STD = {
    "HRV": 935.9717382401759,
    "IR_016": 172.01044433112992,
    "IR_039": 96.53756504807913,
    "IR_087": 96.21369354283686,
    "IR_097": 86.72892737648276,
    "IR_108": 156.20651744208888,
    "IR_120": 104.35287930753246,
    "IR_134": 104.36462050405994,
    "VIS006": 150.2399269307514,
    "VIS008": 152.16086321818398,
    "WV_062": 111.8514878214775,
    "WV_073": 106.8855172848904,
}


class SatelliteML(DataSourceOutputML):
    """Model for output of satellite data"""

    # Shape: [batch_size,] seq_length, width, height, channel
    data: Array = Field(
        ...,
        description="Satellites images. Shape: [batch_size,] seq_length, width, height, channel",
    )
    x: Array = Field(
        ...,
        description="The x (OSGB geo-spatial) coordinates of the satellite images. "
        "Shape: [batch_size,] width",
    )
    y: Array = Field(
        ...,
        description="The y (OSGB geo-spatial) coordinates of the satellite images. "
        "Shape: [batch_size,] height",
    )

    time: Array = Field(
        ...,
        description="Time index of satellite data at 5 minutes past the hour {0, 5, ..., 55}. "
        "*not* the {4, 9, ..., 59} timings of the satellite imagery. "
        "Datetimes become Unix epochs (UTC) represented as int64 just before being"
        "passed into the ML model.",
    )

    channels: Optional[Array] = Field(
        list(SAT_MEAN.keys()), description="List of the satellite channels"
    )

    def get_datetime_index(self) -> Array:
        """Get the datetime index of this data"""
        return self.time

    @staticmethod
    def from_xr_dataset(xr_dataset: xr.Dataset):
        """Change xr dataset to model."""

        # make sure the dims are in the correct order
        xr_dataset = re_order_dims(xr_dataset)

        # convert to torch dictionary
        satellite_batch_ml = xr_dataset.torch.to_tensor(["data", "time", "x", "y"])

        # move to Model
        return SatelliteML(**satellite_batch_ml)

    def normalize(self):
        """Normalize the satellite data"""
        if not self.normalized:
            mean = np.array([SAT_MEAN[b] for b in self.channels])
            std = np.array([SAT_STD[b] for b in self.channels])
            # Need to get to the same shape, so add 3 1-dimensions
            mean = np.expand_dims(mean, axis=[1, 2, 3])
            std = np.expand_dims(std, axis=[1, 2, 3])
            self.data = self.data - mean
            self.data = self.data / std
            self.normalized = True



""" Model for Sun features """
import logging

from nowcasting_dataset.consts import SUN_AZIMUTH_ANGLE, SUN_ELEVATION_ANGLE
from pydantic import Field, validator

from nowcasting_dataloader.data_sources.datasource_output import Array, DataSourceOutputML

logger = logging.getLogger(__name__)


class SunML(DataSourceOutputML):
    """Model for Sun features"""

    sun_azimuth_angle: Array = Field(
        ...,
        description="PV azimuth angles i.e where the sun is. " "Shape: [batch_size,] seq_length",
    )

    sun_elevation_angle: Array = Field(
        ...,
        description="PV elevation angles i.e where the sun is. " "Shape: [batch_size,] seq_length",
    )
    sun_datetime_index: Array

    @validator("sun_elevation_angle")
    def elevation_shape(cls, v, values):
        """
        Validate 'sun_elevation_angle'.

        This is done by change shape is the same as the "sun_azimuth_angle"
        """
        assert v.shape == values["sun_azimuth_angle"].shape
        return v

    @validator("sun_datetime_index")
    def sun_datetime_index_shape(cls, v, values):
        """
        Validate 'sun_datetime_index'.

        This is done by checking last dimension is the same as the last dim of 'sun_azimuth_angle'
        i.e the time dimension
        """
        assert v.shape[-1] == values["sun_azimuth_angle"].shape[-1]
        return v

    def get_datetime_index(self):
        """Get the datetime index of this data"""
        return self.sun_datetime_index

    @staticmethod
    def from_xr_dataset(xr_dataset):
        """Change xr dataset to model. If data does not exist, then return None"""

        sun_batch_ml = xr_dataset.torch.to_tensor(["azimuth", "time", "elevation"])

        sun_batch_ml["sun_datetime_index"] = sun_batch_ml.pop("time")
        sun_batch_ml[SUN_AZIMUTH_ANGLE] = sun_batch_ml.pop("azimuth")
        sun_batch_ml[SUN_ELEVATION_ANGLE] = sun_batch_ml.pop("elevation")

        return SunML(**sun_batch_ml)



""" Model for output of GSP data """
import logging

from pydantic import Field

from nowcasting_dataloader.data_sources.datasource_output import Array, DataSourceOutputML

logger = logging.getLogger(__name__)


class MetadataML(DataSourceOutputML):
    """Model for output of GSP data"""

    batch_size: int = Field(
        ...,
        g=0,
        description="The size of this batch. If the batch size is 0, "
        "then this item stores one data item",
    )

    t0_datetime_utc: Array = Field(
        ...,
        description="The t0s of each example ",
    )

    x_center_osgb: Array = Field(
        ...,
        description="The x centers of each example in OSGB coordinates",
    )

    y_center_osgb: Array = Field(
        ...,
        description="The y centers of each example in OSGB coordinates",
    )



""" Model for Topogrpahic features """
import logging

from nowcasting_dataset.consts import TOPOGRAPHIC_DATA
from pydantic import Field, validator

from nowcasting_dataloader.data_sources.datasource_output import Array, DataSourceOutputML

logger = logging.getLogger(__name__)

# Computed using nowcasting_dataset/notebooks/computs_stats_from_batches.ipynb
# on 2021-11-24 using the training batch of v15 of the dataset.
TOPO_MEAN = 135.88652096070271
TOPO_STD = 145.65013699767726867


class TopographicML(DataSourceOutputML):
    """
    Topographic/elevation map features.
    """

    # Shape: [batch_size,] width, height
    topo_data: Array = Field(
        ...,
        description="Elevation map of the area covered by the satellite data. "
        "Shape: [batch_size], width, height",
    )
    topo_x_coords: Array = Field(
        ...,
        description="The x (OSGB geo-spatial) coordinates of the topographic images. "
        "Shape: [batch_size,] width",
    )
    topo_y_coords: Array = Field(
        ...,
        description="The y (OSGB geo-spatial) coordinates of the topographic images. "
        "Shape: [batch_size,] height",
    )

    @property
    def height(self):
        """The height of the topographic image"""
        return self.topo_data.shape[-1]

    @property
    def width(self):
        """The width of the topographic image"""
        return self.topo_data.shape[-2]

    @validator("topo_x_coords")
    def x_coordinates_shape(cls, v, values):
        """Validate 'topo_x_coords'"""
        assert v.shape[-1] == values["topo_data"].shape[-2]
        return v

    @validator("topo_y_coords")
    def y_coordinates_shape(cls, v, values):
        """Validate 'topo_y_coords'"""
        assert v.shape[-1] == values["topo_data"].shape[-1]
        return v

    @staticmethod
    def from_xr_dataset(xr_dataset):
        """Change xr dataset to model. If data does not exist, then return None"""

        topographic_batch_ml = xr_dataset.torch.to_tensor(["data", "x", "y"])

        topographic_batch_ml[TOPOGRAPHIC_DATA] = topographic_batch_ml.pop("data")
        topographic_batch_ml["topo_x_coords"] = topographic_batch_ml.pop("x")
        topographic_batch_ml["topo_y_coords"] = topographic_batch_ml.pop("y")

        return TopographicML(**topographic_batch_ml)

    def normalize(self):
        """Normalize the topological data"""
        if not self.normalized:
            self.topo_data = self.topo_data - TOPO_MEAN
            self.topo_data = self.topo_data / TOPO_STD
            self.normalized = True



""" Take subsets of xr.datasets """
import logging
from datetime import datetime
from typing import Optional, Union

import numpy as np
import pandas as pd
from nowcasting_dataset.dataset.batch import Batch

logger = logging.getLogger(__name__)


def subselect_data(
    batch: Batch,
    history_minutes: int,
    forecast_minutes: int,
    current_timestep_index: Optional[int] = None,
) -> Batch:
    """
    Subselects the data temporally.

    This function selects all data within the time range [t0-history_minutes, t0+forecast_minutes]

    Args: batch: Example dictionary containing at least the required_keys required_keys: The
    required keys present in the dictionary to use current_timestep_index: The index into either
    SATELLITE_DATETIME_INDEX or NWP_TARGET_TIME giving the current timestep history_minutes: How
    many minutes of history to use forecast_minutes: How many minutes of future data to use for
    forecasting

    Returns:
        Example with only data between [t0 - history_minutes, t0 + forecast_minutes] remaining
    """
    logger.debug(
        f"Select sub data with new historic minutes of {history_minutes} "
        f"and forecast minutes if {forecast_minutes}"
    )

    # We are subsetting the data, so we need to select the t0_dt, i.e the time now for each Example.
    # We in fact only need this from the first example in each batch
    if current_timestep_index is None:
        # t0_dt or if not available use a different datetime index
        t0_dt_of_first_example = pd.Timestamp(batch.metadata.t0_dt[0].values[0])
    else:
        if batch.satellite is not None:
            t0_dt_of_first_example = batch.satellite.time[0, current_timestep_index].values
        elif batch.hrvsatellite is not None:
            t0_dt_of_first_example = batch.hrvsatellite.time[0, current_timestep_index].values
        elif batch.pv is not None:
            t0_dt_of_first_example = batch.pv.time[0, current_timestep_index].values
        else:
            t0_dt_of_first_example = batch.nwp.time[0, current_timestep_index].values

    if batch.satellite is not None:
        batch.satellite = select_time_period(
            x=batch.satellite,
            history_minutes=history_minutes,
            forecast_minutes=forecast_minutes,
            t0_dt_of_first_example=t0_dt_of_first_example,
        )

    # Now for NWP, if used
    if batch.nwp is not None:
        batch.nwp = select_time_period(
            x=batch.nwp,
            history_minutes=history_minutes,
            forecast_minutes=forecast_minutes,
            t0_dt_of_first_example=t0_dt_of_first_example,
        )

    # Now for GSP, if used
    if batch.gsp is not None:
        batch.gsp = select_time_period(
            x=batch.gsp,
            history_minutes=history_minutes,
            forecast_minutes=forecast_minutes,
            t0_dt_of_first_example=t0_dt_of_first_example,
        )

    # Now for PV, if used
    if batch.pv is not None:
        batch.pv = select_time_period(
            x=batch.pv,
            history_minutes=history_minutes,
            forecast_minutes=forecast_minutes,
            t0_dt_of_first_example=t0_dt_of_first_example,
        )

    # Now for SUN, if used
    if batch.sun is not None:
        batch.sun = select_time_period(
            x=batch.sun,
            history_minutes=history_minutes,
            forecast_minutes=forecast_minutes,
            t0_dt_of_first_example=t0_dt_of_first_example,
        )

    # DATETIME TODO

    return batch


def select_time_period(
    x,
    history_minutes: int,
    forecast_minutes: int,
    t0_dt_of_first_example: Union[datetime, pd.Timestamp],
):
    """
    Selects a subset of data between the indicies of [start, end] for each key in keys

    Note that class is edited so nothing is returned.

    Args:
        x: dataset that is ot be reduced
        t0_dt_of_first_example: datetime of the current time (t0) in the first example of the batch
        history_minutes: How many minutes of history to use
        forecast_minutes: How many minutes of future data to use for forecasting

    """
    logger.debug(
        f"Taking a sub-selection of the batch data based on a history minutes of {history_minutes} "
        f"and forecast minutes of {forecast_minutes}"
    )

    start_time_of_first_example = t0_dt_of_first_example - pd.to_timedelta(
        f"{history_minutes} minute 30 second"
    )
    end_time_of_first_example = t0_dt_of_first_example + pd.to_timedelta(
        f"{forecast_minutes} minute 30 second"
    )

    logger.debug(f"New start time for first example is {start_time_of_first_example}")
    logger.debug(f"New end time for first example is {end_time_of_first_example}")

    if hasattr(x, "time"):
        time_of_first_example = pd.to_datetime(x.time[0].values)

    else:
        # for nwp, maybe rename
        time_of_first_example = pd.to_datetime(x.target_time[0].values)

    # find the start and end index, that we will then use to slice the data
    start_i, end_i = np.searchsorted(
        time_of_first_example, [start_time_of_first_example, end_time_of_first_example]
    )

    # slice all the data
    y = x.where(((x.time_index >= start_i) & (x.time_index < end_i)), drop=True)

    # the `where` statement seems to add extra dimension to data variables,
    # if they don't originally have `time_index` dimension
    for key in y.keys():
        if "time_index" not in getattr(x, key).dims:
            y.__setitem__(key, getattr(x, key))

    return y



""" A class to create a fake dataset """
import torch
from nowcasting_dataset.config.model import Configuration
from nowcasting_dataset.dataset.batch import Batch

from nowcasting_dataloader.batch import BatchML
from nowcasting_dataloader.utils.position_encoding import generate_position_encodings_for_batch


class FakeDataset(torch.utils.data.Dataset):
    """Fake dataset."""

    def __init__(
        self,
        configuration: Configuration,
        length: int = 10,
        add_position_encoding: bool = False,
    ):
        """
        Init

        Args:
            configuration: configuration object
            length: length of dataset
            add_position_encoding: Whether to add position encoding or not
        """
        self.add_position_encoding = add_position_encoding
        self.number_nwp_channels = len(configuration.input_data.nwp.nwp_channels)
        self.length = length
        self.configuration = configuration

    def __len__(self):
        """Number of pieces of data"""
        return self.length

    def per_worker_init(self, worker_id: int):
        """Nothing to do for FakeDataset"""
        pass

    def __getitem__(self, idx):
        """
        Get item, use for iter and next method

        Args:
            idx: batch index

        Returns: Dictionary of random data

        """
        if self.add_position_encoding:
            batch: Batch = Batch.fake(configuration=self.configuration)
            position_encodings = generate_position_encodings_for_batch(batch, num_bands=12)
            batch: BatchML = BatchML.from_batch(batch=batch)
            batch: dict = batch.dict()
            batch.update(position_encodings)
            return batch
        else:
            x: BatchML = BatchML.fake(configuration=self.configuration)
            return x.dict()



"""
This file contains various ways of performing positional encoding.

These encodings can be:
- Absolute positioning (i.e. this pixel is at this latitude/longitude, and is at 16:00)

These encodings can also be performed with:
- Fourier Features, based off what is done in PerceiverIO
"""
import datetime
from math import pi
from typing import Dict, List, Optional, Tuple

import einops
import numpy as np
import pandas as pd
import torch
import xarray as xr
from nowcasting_dataset.dataset.batch import Batch
from nowcasting_dataset.geospatial import lat_lon_to_osgb

TIME_DIM = 2
HEIGHT_DIM = 3
WIDTH_DIM = 4
# For GSP and PV, have an ID dimension
ID_DIM = 3


def get_seviri_rss_bounds() -> Dict[str, float]:
    """
    Computes the SEVIRI RSS bounds in OSGB coordinates

    SEVIRI RSS is the imager that takes all the satellite imagery currently used by
    the nowcasting dataset and models

    Returns:
        Dictionary containing the geographic bounds of the RSS images in OSGB coordinates
    """
    x_min = np.inf
    x_max = -np.inf
    y_min = np.inf
    y_max = -np.inf
    for lat in [15, 70]:
        for lon in [-45, 65]:
            x, y = lat_lon_to_osgb(lat, lon)
            if x < x_min:
                x_min = x
            if x > x_max:
                x_max = x
            if y < y_min:
                y_min = y
            if y > y_max:
                y_max = y
    return {
        "x_min": x_min,
        "y_min": y_min,
        "x_max": x_max,
        "y_max": y_max,
    }


def generate_position_encodings_for_batch(
    batch: Batch, time_range: Optional[Tuple[datetime.datetime, datetime.datetime]] = None, **kwargs
) -> dict[str, torch.Tensor]:
    """
    Generates positional encodings and returns them as a dictionary

    This is not returned with the Batch, as that would require more keys, etc. in Batch

    Args:
        batch: Batch object holding the data
        time_range: List of start_year, end_year datetimes to normalize against,
            time_range[0].year must be less than time_range[1].year

    Returns:
        Dictionary containing the keys of the modalities in the Batch + '_position_encoding'
    """

    position_encodings = {}
    # Go for each modality where a position encoding makes sense
    for k in batch.__fields__.keys():
        if k in [
            "nwp",
            "satellite",
            "hrvsatellite",
            "topographic",
            "gsp",
            "pv",
        ]:
            xr_dataset = getattr(batch, k)
            if xr_dataset is not None:
                datetimes = None
                if hasattr(xr_dataset, "time"):
                    datetimes = xr_dataset.time.values
                geospatial_coordinates = (
                    [xr_dataset.x.values, xr_dataset.y.values]
                    if "x_index" in xr_dataset.sizes
                    else [xr_dataset.x_coords.values, xr_dataset.y_coords.values]
                )
                position_encodings[k + "_position_encoding"] = encode_absolute_position(
                    shape=determine_shape_of_encoding(xr_dataset),
                    geospatial_coordinates=geospatial_coordinates,
                    datetimes=datetimes,
                    geospatial_bounds=get_seviri_rss_bounds(),
                    time_range=time_range,
                    **kwargs,
                )

    return position_encodings


def determine_shape_of_encoding(xr_dataset: xr.Dataset) -> List[int]:
    """
    Determine the shape of the encoding needed for the batch example

    Args:
        xr_dataset: Xarray dataset containing the data

    Returns:
        The determined shape that the encoding needs to be, either a 5-element list
        for image modalities, or a 4-element list for point modalities (GSP, PV systems)
    """
    channel_key = "channels_index" if "channels_index" in xr_dataset.sizes else "id_index"
    shape = [xr_dataset.sizes["example"]]
    shape.append(
        xr_dataset.sizes.get(channel_key, 1)
    )  # If no channels, count as single channel image)
    shape.append(xr_dataset.sizes.get("time_index", 1))  # If no time dimension, just a single one)

    # Now for the main issue, either 4 or 5D here
    if "x_index" in xr_dataset.sizes:
        shape.append(xr_dataset.sizes["x_index"])
        shape.append(xr_dataset.sizes["y_index"])
    else:
        # No spatial extant i.e. GSP, or PV
        # Then the output should be for each ID, so would then be the same as the channels ID
        shape.append(xr_dataset.sizes["id_index"])
    return shape


def encode_modalities(
    modalities_to_encode: Dict[str, torch.Tensor],
    datetimes: Dict[str, List[datetime.datetime]],
    geospatial_coordinates: Dict[str, Tuple[np.ndarray, np.ndarray]],
    geospatial_bounds: Dict[str, float],
    time_range: Optional[Tuple[datetime.datetime, datetime.datetime]] = None,
    **kwargs,
) -> dict[str, torch.Tensor]:
    """
    Create a consistent position encoding and encode the positions of the different modalities

    This position encoding is added as new keys to the dictionary containing the modalities
    to encode. This is done instead of appending the position encoding in case the position
    encoding needs to be used for the query to the
    Perceiver IO model

    This code assumes that there is at least 2 timesteps of at least one modality to be encoded

    Args:
        modalities_to_encode: Dict of input modalities, i.e. NWP, Satellite, PV, GSP, etc as torch.
            Tensors in [B, C, T, H, W] ordering
        datetimes: Dict of datetimes for each modality, giving the actual date for each timestep in
            the modality
        geospatial_coordinates: Dict of x, y coordinates for each modality with pixels, used to
            determine smallest spatial step needed, in OSGB coordinates
        geospatial_bounds: Max extant of the area where examples could be drawn from, used for
            normalizing coordinates within an area of interest
            in the format of a dictionary with the keys {'x_min', 'x_max', 'y_min', 'y_max'}
        time_range: List of start_year, end_year datetimes to normalize against,
            time_range[0].year must be less than time_range[1].year
        kwargs: Passed to fourier_encode

    Returns:
        Input modality dictionary where for every 'key' in modalities_to_encode, a new key called
            'key+'_position_encoding' will be added
        containing the absolute position encoding of the examples
    """
    position_encodings = {}
    for key in modalities_to_encode.keys():
        position_encodings[key + "_position_encoding"] = encode_absolute_position(
            shape=modalities_to_encode[key].shape,
            geospatial_coordinates=geospatial_coordinates[key],
            datetimes=datetimes[key],
            geospatial_bounds=geospatial_bounds,
            time_range=time_range,
            **kwargs,
        )
    # Update original dictionary
    modalities_to_encode.update(position_encodings)
    return modalities_to_encode


def encode_absolute_position(
    shape: List[int],
    geospatial_coordinates: List[np.ndarray],
    geospatial_bounds: Dict[str, float],
    datetimes: Optional[List[datetime.datetime]] = None,
    time_range: Optional[Tuple[datetime.datetime, datetime.datetime]] = None,
    num_bands: int = 4,
) -> torch.Tensor:
    """
    Encodes the absolute position of the pixels/voxels in time and space

    This should be done per-modality and can be thought of as the relative position of the
    input modalities across a given year and the area of the Earth covered by all the examples.

    Args:
        shape: Shape to encode positions for
        geospatial_coordinates: The geospatial coordinates, in OSGB format
        datetimes: Time of day and date as a list of datetimes, one for each timestep
        geospatial_bounds: The geospatial bounds of the area where the examples come from, e.g.
            the coordinates of the area covered by the SEVIRI RSS image
        time_range: List of start_year, end_year datetimes to normalize against,
            time_range[0].year must be less than time_range[1].year
        num_bands: Number of bands to pass to Fourier encoding

    Returns:
        The absolute position encoding for the given shape
    """
    # Fourier Features of absolute position
    encoded_geo_position = normalize_geospatial_coordinates(
        geospatial_coordinates,
        geospatial_bounds,
        max_freq=(2 * len(geospatial_coordinates) + 1),
        num_bands=num_bands,
    )
    absolute_position_encoding = einops.repeat(
        encoded_geo_position, "b h w c -> b c t h w", t=shape[TIME_DIM]
    )

    if len(shape) == 4:  # Point systems
        # Probably GSP or PV, just need the diagonal of the values, so have B, C, T, GSP/PV ID
        absolute_position_encoding = torch.diagonal(absolute_position_encoding, dim1=-2, dim2=-1)

    if datetimes is not None:
        datetime_features = create_datetime_features(datetimes)
        if time_range is not None:
            # Add year feature
            year_features = encode_year(datetimes, time_range)
            # Repeat
            year_features = einops.repeat(
                year_features, "b t c -> b (repeat t) c", repeat=shape[TIME_DIM]
            )
            datetime_features.append(year_features)
        absolute_position_encoding = combine_space_and_time_features(
            absolute_position_encoding, datetime_features=datetime_features, shape=shape
        )
    return absolute_position_encoding


def combine_space_and_time_features(
    spatial_features: torch.Tensor, datetime_features: List[torch.Tensor], shape: List[int]
) -> torch.Tensor:
    """
    Combine spatial and temporal features a list of Tensors to be concatenated

    Args:
        spatial_features: Spatial features
        datetime_features: List of datetime features
        shape: The desired shape of the encoding

    Returns:
        Tensor containing the combined space and time features
    """
    to_concat = [spatial_features]
    # Combine time and space features
    for date_feature in datetime_features:
        if len(shape) == 5:
            date_feature = einops.repeat(
                date_feature, "b t c -> b c t h w", h=shape[HEIGHT_DIM], w=shape[WIDTH_DIM]
            )
        else:
            date_feature = einops.repeat(date_feature, "b t c -> b c t id", id=shape[ID_DIM])
        to_concat.append(date_feature)
    space_and_time_encoding = torch.cat(to_concat, dim=1)
    return space_and_time_encoding


def normalize_geospatial_coordinates(
    geospatial_coordinates: List[np.ndarray], geospatial_bounds: Dict[str, float], **kwargs
) -> torch.Tensor:
    """
    Normalize the geospatial coordinates by the max extant to keep everything between -1 and 1

    This normalization should be against a set geospatial area, so that the same place has the
     same spatial encoding every time.

    Args:
        geospatial_coordinates: The coordinates for the pixels in the image
        geospatial_bounds: The maximum extant

    Returns:
        The normalized geospatial coordinates, rescaled to between -1 and 1 for the whole extant of
         the training area

    """
    # Normalize the X first
    geospatial_coordinates[0] = (geospatial_coordinates[0] - geospatial_bounds["x_min"]) / (
        geospatial_bounds["x_max"] - geospatial_bounds["x_min"]
    )
    # Normalize the Y second
    geospatial_coordinates[1] = (geospatial_coordinates[1] - geospatial_bounds["y_min"]) / (
        geospatial_bounds["y_max"] - geospatial_bounds["y_min"]
    )

    # Now those are between 0 and 1, want between -1 and 1
    geospatial_coordinates[0] = geospatial_coordinates[0] * 2 - 1
    geospatial_coordinates[1] = geospatial_coordinates[1] * 2 - 1
    # Now create a grid of the coordinates
    # Have to do it for each individual example in the batch, and zip together x and y for it
    solid_tensor = None
    for idx in range(len(geospatial_coordinates[0])):
        x = torch.from_numpy(geospatial_coordinates[0][idx])
        y = torch.from_numpy(geospatial_coordinates[1][idx])
        grid = torch.meshgrid(x, y)
        pos = torch.stack(grid, dim=-1)
        encoded_position = fourier_encode(pos, **kwargs)
        encoded_position = einops.rearrange(encoded_position, "... n d -> ... (n d)")
        if solid_tensor is None:
            solid_tensor = torch.zeros(
                len(geospatial_coordinates[0]), *encoded_position.size(), dtype=x.dtype
            )
        solid_tensor[idx] = encoded_position

    return solid_tensor


def encode_year(
    datetimes: List[List[datetime.datetime]],
    time_range: Tuple[datetime.datetime, datetime.datetime],
    num_bands: int = 4,
) -> torch.Tensor:
    """
    Encode the year of each example in the batch, normalizing between the start and end date

    Args:
        datetimes: The datetimes for all the examples in the batch, in format
            dateimes[batch_idx][timestep_idx]
        time_range: List of start_year, end_year datetimes to normalize against,
            time_range[0].year must be less than time_range[1].year
        num_bands: Number of bands to include in the Fourier encoding

    Returns:
        Tensor containing the encoding for the year of the first timestep in each example
            This should be correct as each example covers a tiny part of a year, and since
            we do not predict at night, there should be no examples covering two years
    """

    assert time_range[0].year < time_range[1].year, (
        "First datetime must have a lower year than " "the second"
    )

    year_encoding = []
    for batch_idx in range(len(datetimes)):
        encoding = (pd.Timestamp(datetimes[batch_idx][0]).year - time_range[0].year) / (
            time_range[1].year - time_range[0].year
        )
        # Rescale between -1 and 1
        encoding *= 2
        encoding -= 1
        # Compute Fourier Features
        encoding = fourier_encode(torch.as_tensor([encoding]), max_freq=99, num_bands=num_bands)
        year_encoding.append(encoding)
    year_encoding = torch.stack(year_encoding, dim=0)
    return year_encoding


def create_datetime_features(
    datetimes: List[List[datetime.datetime]], num_bands: int = 4
) -> List[torch.Tensor]:
    """
    Converts a list of datetimes to day of year, hour of day sin and cos representation

    Args:
        datetimes: List of list of datetimes for the examples in a batch
        num_bands: Number of bands to include in the Fourier encoding

    Returns:
        Tuple of torch Tensors containing the hour of day sin,cos, and day of year sin,cos
    """
    hour_of_day = []
    day_of_year = []
    for batch_idx in range(len(datetimes)):
        hours = []
        days = []
        for index in datetimes[batch_idx]:
            time_index = pd.Timestamp(index)
            hours.append((((time_index.hour + (time_index.minute / 60)) / 24) * 2) - 1)
            days.append(((time_index.timetuple().tm_yday / 366) * 2) - 1)
        hour_of_day.append(hours)
        day_of_year.append(days)

    outputs = []
    hour_of_day = torch.as_tensor(hour_of_day)
    day_of_year = torch.as_tensor(day_of_year)
    # Compute Fourier Features
    hour_of_day = fourier_encode(hour_of_day, max_freq=49, num_bands=num_bands)
    day_of_year = fourier_encode(day_of_year, max_freq=733, num_bands=num_bands)
    outputs.append(hour_of_day)
    outputs.append(day_of_year)

    return outputs


def fourier_encode(
    x: torch.Tensor,
    max_freq: float,
    num_bands: int = 4,
    sine_only: bool = False,
) -> torch.Tensor:
    """
    Create Fourier Encoding

    Args:
        x: Input Torch Tensor
        max_freq: Maximum frequency for the Fourier features
        num_bands: Number of frequency bands
        sine_only: Whether to only use sine or both sine and cosine features

    Returns:
        Torch Tensor with the fourier position encoded concatenated
    """
    x = x.unsqueeze(-1)
    device, dtype, orig_x = x.device, x.dtype, x

    scales = torch.linspace(
        16.0,
        max_freq / 2,
        num_bands,
        device=device,
        dtype=dtype,
    )
    scales = scales[(*((None,) * (len(x.shape) - 1)), Ellipsis)]

    x = x * scales * pi
    if sine_only:
        x = torch.cat((x.sin(), orig_x), dim=-1)
    else:
        x = torch.cat((x.sin(), x.cos(), orig_x), dim=-1)
    return x
